--------------------------------------------------------------------
-- Return to Castle Wolfenstein MDC Importer for gmax
-- Copyright (c) Chris Cookson, 2001
--
-- cjcookson@hotmail.com
--
-- For gmax Scripts, News, Articles and More check out:
--
-- http://www.gmaxsupport.com and http://mojo.gmaxsupport.com
--
-- Thanks to Ryan Feltrin and Hewster
-- 
-- Version history:
--  1.1a Tweaks by Hewster
--	1.1  Tag animation support
-- 	1.0  First public release
--
--------------------------------------------------------------------


utility MDCImport "RtCW MDC Import" silentErrors:false
(
	-- Game data path is location of unzipped .PK3 files
	global g_gameDataPath = ""
	global g_defaultGameDataPath = "c:\\program files\\Return to Castle Wolfenstein\\models\\Main\\Data"
	
	-- Useful MDC file information
	local modelFilename
	local modelName
	local modelPath

	local appName = "Wolfenstein MDC Importer"

	--------------------
	-- User Interface --
	--------------------

	group "About:"
	(
		label titleLabel	"Wolfenstein MDC Import v1.1a"
		HyperLink addy "by Chris Cookson" align:#center address:"mailto:cjcookson@hotmail.com" color:(color 0 0 170) hoverColor:(color 170 0 0)
	)
	
	group "Options:"
	(
		edittext pakDirEdit "Data Path:" text:"** Not Set **" readonly:true
		button setPakRootButton "Set Game Data Path"
	)
	
	group "Import:"
	(
		radiobuttons tagAction "Tags:" labels:#("Ignore all", "Load all", "Auto attach") default:1

		checkbox promptForMissing "Prompt for missing textures" checked:true

		
		button importButton "Import MDC..."	
	)
	
	-----------------------------------------------------
	-- Recent file list (sadly I can't save to an INI) --
	-----------------------------------------------------

	local recentFile

	fn AddRecentFile fname=
	(
		recentFile = fname
		ok
	)
	
	fn GetMostRecentPath=
	(
		if recentFile != undefined then
		(
			return getFileNamePath recentFile
		) else if g_gameDataPath != undefined then
		(
			return g_gameDataPath + "\\models\\"
		) else
		(
			return ""
		)
	)

	---------------
	-- Utilities --
	---------------

	-- Replacement for doesFileExist which seems to give false positives
	fn DoesFileReallyExist filename=
	(
		local temp = fopen filename "rb"
		if temp != undefined then
		(
			fclose temp
			true
		)
		else
		(
			false
		)
	)
	
	-- Error message box
	fn ShowError msg=
	(
		format "*** Error: %\n" msg
		messageBox msg title:appName;
		ok
	)
	
	-- Error message box and exception
	fn FatalError msg=
	(
		ShowError msg;
		throw (appName + ":" + msg)
	)
	
	-- Quick check, error if fail
	fn Check condition msg =
	(
		if not condition do
		(
			if msg == unsupplied do msg = "Check failed"
			format "*** Check failed: %\n" msg
			FatalError msg
		)
	)
	
	-- Binstream tools
	
	-- Skip over a number of bytes
	fn SkipBytes bstream count=
	(
		local unknown
		case count of
		(
			2: unknown = ReadShort bstream #unsigned
			4: unknown = ReadLong bstream #unsigned
			default:
			(
				for i = 1 to count do
				(
					unknown = ReadByte bstream #unsigned
				)
			)
		)
	)
	

	-- Read a string of fixedLen chars
	fn ReadFixedString bstream fixedLen=
	(
		local str = ""
		for i = 1 to fixedLen do
		(
			local ch = ReadByte bstream #unsigned
			str += bit.intAsChar ch
			if ch == 0 then
			(
				-- Ignore junk
				SkipBytes bstream (fixedLen - i)
				exit
			)
		)
		str
	)
	
	-- String utils
	
	-- Get the string a long represents
	fn LongToString num=
	(
		local str = ""
		for i = 1 to 4 do
		(
			str += bit.intAsChar (bit.and num 0xff)
			-- num = bit.shift num -8
			num /= 256
		)
		str
	)
	
	-- Convert to lowercase
	fn ToLowerCase instr=
	(
		local outstr = copy instr
		local upper = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
		local lower = "abcdefghijklmnopqrstuvwxyz"
		
		for i =1 to outstr.count do
		(
			local pos = findString upper outstr[i]
			if pos != undefined do outstr[i] = lower[pos]
		)
		outstr

	)
	
	
	-- Convert a forward-slash style PK3 path to a windows one
	fn ToWindowsPath pakPath=
	(
		local winPath = "\\"
		for i = 1 to pakPath.count do
		(
			if pakPath[i] == "/" then
			(
				winPath += "\\"
			) else
			(
				winPath += pakPath[i]
			)
		)
		winPath
	)
	
	-- Swap a file's extension	
	fn ChangeExt filename newExt=
	(
		(getFilenamePath filename) + (getFilenameFile filename) + newExt
	)
	
	-- Get the game data path set by the user, or a default one
	fn GetDataPath=
	(
		if g_gameDataPath != undefined then
		(
			g_gameDataPath
		)
		else
		(
			g_defaultGameDataPath
		)
	)
	
	-- Extract the filename of a data file from a PK3 filename
	fn GetDataPathFilename filename=
	(
		local splitPath = filterString filename "\\/"
		
		if splitPath.count > 0 then
		(
			splitPath[splitPath.count]
		)
		else
		(
			""
		)
	)
	
	-- Cache texture names so we don't keep prompting the user for the same ones
	struct TextureCache
	(
		textureNames,
		textures,
		
		-- Clear the cache (do this per model)
		fn Clear=
		(
			textureNames = #()
			textures = #()
		),
		
		-- Add a texture to the cache and return the loaded bitmaptexture
		fn AddTexture name fname=
		(
			local bmpTex = bitmapTexture filename:fname
			append textureNames name
			append textures bmpTex
			bmpTex
		),
		
		-- Find a texture given a PK3 style filename
		fn FindTexture filename=
		(
			-- First of all, check the cache
			local index = findItem textureNames filename
			if index != 0 then
			(
				return textures[index]
			)
		
			-- Not in cache so we'll try searching in the model's directory
			local localFilename = modelPath + (GetDataPathFilename filename)
			
			--format "Not in cache, looking for %\n" localFilename
			
			if DoesFileReallyExist localFilename then
			(
				return (AddTexture filename localFilename)
			)
			
			-- Try different file extensions / image types (MDC files often lie :)		
			if (ToLowerCase (getFileNameType localFilename)) == ".jpg" then
			(
				local tgaFilename = ChangeExt localFilename ".tga"
				
				if DoesFileReallyExist tgaFilename then
				(
					format "Warning: Expected %, but using %\n" localFilename tgaFilename
					
					return (AddTexture filename tgaFilename)
				)
			)
			else
			(
				local jpgFilename = ChangeExt localFilename ".jpg"
				--format "Trying %\n" jpgFilename
				if DoesFileReallyExist jpgFilename then
				(
					--format "Warning: Expected %, but using %\n" localFilename jpgFilename
					return (AddTexture filename jpgFilename)
				)
			)
			
			-- No luck, try looking in the unzipped game data path
			local dataFilename = GetDataPath() + (ToWindowsPath filename)
	
			if DoesFileReallyExist dataFilename then
			(
				return (AddTexture filename dataFilename)
			)
			
			-- Try different file extensions / image types (MDC files often lie :)		
			if (ToLowerCase (getFileNameType dataFilename)) == ".jpg" then
			(
				local tgaFilename = ChangeExt dataFilename ".tga"
				
				if DoesFileReallyExist tgaFilename then
				(
					format "Warning: Expected %, but using %\n" dataFilename tgaFilename
					return (AddTexture filename tgaFilename)
				)
			)
			else
			(
				local jpgFilename = ChangeExt dataFilename ".jpg"
				if DoesFileReallyExist jpgFilename then
				(
					format "Warning: Expected %, but using %\n" dataFilename jpgFilename
					return (AddTexture filename jpgFilename)
				)
			)
			
			-- We're out of luck so ask the user to find it for us!
			local foundFilename
			if promptForMissing.checked do
			(
				messageBox ("Can't find texture " + filename + "\nPlease click OK and locate it.")
			
				foundFilename = getOpenFileName caption:("Find Texture " + filename) filename:localFilename \
							types:"All supported (*.tga;*.jpg;*.jpeg)|*.tga;*.jpg;*.jpeg|All Files (*.*)|*.*|"
			)
	
			if foundFilename != undefined then
			(
				return (AddTexture filename foundFilename)
			)
			else
			(
				format "Warning: Could not find % or any variant!\n" filename
				append textureNames filename
				append textures undefined
			)
			undefined
		)
	)


	-------------
	-- UI Bits --
	-------------
	
	fn UpdateUI=
	(
		local theDir = g_gameDataPath
		if theDir == undefined do
		(
			theDir = "** Not Set **"
		)
		pakDirEdit.text = theDir
	
	)
		
	on MDCImport open do
	(
		-- Read config from .INI - not much point if I can't write one :(
		/*
		local iniFile = (GetDir #scripts) + "\\mdcimport.ini"
		g_gameDataPath = getINISetting iniFile "Directories" "DataPath"
		*/
		
		if g_gameDataPath == "" then
		(
			g_gameDataPath = g_defaultGameDataPath
		)
	
		UpdateUI()
	)
	
	on MDCImport close do
	(
		-- Write config to .INI ARGH! Not supported in gmax!
		/*
		local iniFile = (GetDir #scripts) + "\\mdcimport.ini"
		setINISetting iniFile "Directories" "DataPath" g_gameDataPath
		*/
	)

	
	--------------------------------------------------------------------
	-- MDC Importer
	--------------------------------------------------------------------
	
	-- Surfaces are a group of polys with the same rendering style
	struct MDC_Surface
	(
		ident,
		name,
		flags,
		numCompFrames,
		numBaseFrames,
		numShaders,
		numVertices,
		numTriangles,
		offsetTriangles,
		offsetShaders,
		offsetTexCoords,
		offsetVertices,
		offsetCompVerts,
		offsetFrameBaseFrames,
		offsetFrameCompFrames,
		offsetEnd,
		
		-- Internal
		offsetStart,	-- Set when LoadFromStream called
		
		fn LoadFromStream bstream=
		(
			offsetStart = ftell bstream
			
			ident					= ReadLong bstream #unsigned
			name					= ReadFixedString bstream 64
			flags					= ReadLong bstream #unsigned
			numCompFrames			= ReadLong bstream #unsigned
			numBaseFrames			= ReadLong bstream #unsigned
			numShaders				= ReadLong bstream #unsigned
			numVertices				= ReadLong bstream #unsigned
			numTriangles			= ReadLong bstream #unsigned
			offsetTriangles			= ReadLong bstream #unsigned
			offsetShaders			= ReadLong bstream #unsigned
			offsetTexCoords			= ReadLong bstream #unsigned
			offsetVertices			= ReadLong bstream #unsigned
			offsetCompVerts			= ReadLong bstream #unsigned
			offsetFrameBaseFrames	= ReadLong bstream #unsigned
			offsetFrameCompFrames	= ReadLong bstream #unsigned
			offsetEnd				= ReadLong bstream #unsigned
		),
		
		fn IsValid=
		(
			if ident != 0x07 then
			(
				return false
			)
			true
		),
		
		fn DebugOut=
		(
			format "MDC_Surface\n-----------\n"
			format "ident = %, name = %, flags = %\n" ident name flags
			format "numCompFrames = %, numBaseFrames = %, numShaders = %\n" numCompFrames numBaseFrames numShaders
			format "numVertices = %, numTriangles = %\n" numVertices numTriangles
			format "ofsTris = %, ofsShaders = %, ofsTexCoords = %, ofsVertices = %, ofsBaseFrames = %, ofsCompFrames = %\n"	\
				offsetTriangles offsetShaders offsetTexCoords offsetVertices\
				offsetFrameBaseFrames offsetFrameCompFrames
		)
	)
	
	-- Triangle defined as 3 indices
	struct MDC_Triangle
	(
		indices,
		
		fn LoadFromStream bstream=
		(
			local v1 = ReadLong bstream #unsigned
			local v2 = ReadLong bstream #unsigned
			local v3 = ReadLong bstream #unsigned
			indices = [v3 + 1, v2 + 1, v1 + 1]
		)
	)
	
	-- Vertex definition
	struct MDC_Vertex
	(
		position,
		normal,
		
		fn LoadFromStream bstream=
		(
			local x = ReadShort bstream #signed
			local y = ReadShort bstream #signed
			local z = ReadShort bstream #signed
			position = [x as float, y as float, z as float] / 64
			normal = ReadShort bstream #unsigned
		),
		
		fn GetSize= 8
	)
	
	-- Texture co-ordinate
	struct MDC_TexCoord
	(
		stu,
		
		fn LoadFromStream bstream=
		(
			local s = ReadFloat bstream
			local t = ReadFloat bstream
			-- gmax needs the t co-ordinate reversed
			stu = [s, -t, 0.0]
		)
	)
	
	-- Shader definition
	struct MDC_Shader
	(
		name,
		flags,
		
		fn LoadFromStream bstream=
		(
			name = ReadFixedString bstream 64
			flags = ReadLong bstream #unsigned
		)
	)
		
	-- Tag name
	struct MDC_TagName
	(
		name,
		
		fn LoadFromStream bstream=
		(
			name = ReadFixedString bstream 64
		)
	)

	
	-- Tag frame: specifies orientation and position of a tag
	struct MDC_TagFrame
	(
		xyz,
		angles,
		
		-- Internal
		isRelative,
		matrix,
		
		fn LoadFromStream bstream=
		(
			local x = ReadShort bstream #signed
			local y = ReadShort bstream #signed
			local z = ReadShort bstream #signed
			local Hx = ( x >= -0.1 and x <= 0.1 )
			local Hy = ( y >= -0.1 and y <= 0.1 )
			local Hz = ( z >= -0.1 and z <= 0.1 )
			isRelative = ( Hx and Hy and Hz )			-- Hewster adjusted so not so critical
			xyz = [x as float, y as float, z as float] / 64
			x = ReadShort bstream #signed
			y = ReadShort bstream #signed
			z = ReadShort bstream #signed
			angles = [x as float, y as float, z as float] * (360.0 / 32767.0)
			
			matrix = (eulerAngles angles.z angles.x angles.y) as matrix3
			matrix.row4 = xyz
		),
		
		fn GetRotation=
		(
			matrix.rotationpart
		),
		
		fn GetMatrix=
		(
			matrix
		)
		
	)
	
	-- Frame: defines basic frame properties
	struct MDC_Frame
	(
		bboxMin,
		bboxMax,
		localOrigin,
		radius,
		name,			-- often creator
		
		fn LoadFromStream bstream=
		(
			local x = ReadFloat bstream
			local y = ReadFloat bstream
			local z = ReadFloat bstream
			bboxMin = [x, y, z]
			x = ReadFloat bstream
			y = ReadFloat bstream
			z = ReadFloat bstream
			bboxMax = [x, y, z]
			x = ReadFloat bstream
			y = ReadFloat bstream
			z = ReadFloat bstream
			localOrigin = [x, y, z]
			radius = ReadFloat bstream
			name = ReadFixedString bstream 16
		)		
	)
	
	-- Header: found at start of file
	struct MDC_Header
	(
		ident,
		version,
		name,
		flags,
		numFrames,
		numTags,
		numSurfaces,
		numSkins,
		offsetFrames,
		offsetTagNames,
		offsetTags,
		offsetSurfaces,
		offsetEnd,
		
		fn LoadFromStream bstream=
		(
			ident			= ReadLong bstream #unsigned
			version			= ReadLong bstream #unsigned
			name			= ReadFixedString bstream 64
			flags			= ReadLong bstream #unsigned
			numFrames		= ReadLong bstream #unsigned
			numTags			= ReadLong bstream #unsigned
			numSurfaces		= ReadLong bstream #unsigned
			numSkins		= ReadLong bstream #unsigned
			offsetFrames	= ReadLong bstream #unsigned
			offsetTagNames	= ReadLong bstream #unsigned
			offsetTags		= ReadLong bstream #unsigned
			offsetSurfaces	= ReadLong bstream #unsigned

			offsetEnd		= ReadLong bstream #unsigned
		),

		fn IsValid=
		(
			local identStr = LongToString ident
			if (LongToString ident) != "IDPC" then return false
			if version != 0x2 then
			(
				ShowError "Incorrect version! Found " + version + " but should be 2"
				return false
			)
			if numFrames < 1 then
			(
				ShowError "MDC does not have any frames!"
				return false
			)
			true
		),
		
		fn DebugOut=
		(
			format "MDC_Header\n----------\n"
			format "Ident: %, version: %, name: %\n" ident version name
			format "Flags: %, numFrames: %, numTags: %, numSurfaces: %, numSkins: %\n" flags numFrames numTags numSurfaces numSkins
			format "ofsFrames: %, ofsTagNames: %, ofsTags: %, ofsSurfs: %, ofsEnd: %\n" offsetFrames offsetTagNames offsetTags offsetSurfaces offsetEnd
			ok
		)
	)	
	

	-- The main MDC loader object
	struct MDC_Loader
	(
		bstream,
		
		header,	
		surfaces,
		tagFrames,
		tagNames,
		tagObjects,
		
		frames,
		
		texCache,
			
		fn Open filename=
		(
			modelFilename = filename
			modelName = getFilenameFile filename
			modelPath = getFileNamePath filename
			bstream = fopen filename "rb"
			
			texCache = TextureCache()
			texCache.Clear()
			ok
		),
		
		fn Close=
		(
			fclose bstream
			ok
		),
		
		-- Create a tag object in the form of the standard Quake3 tag triangle
		fn CreateTagObject name mdcTag=
		(
			local verts = #([0, 0, 0], [0, -1, 0], [2, 0, 0])
			local tri = #([1, 2, 3])
			local tagObject = mesh name:name vertices:verts faces:tri pos:(mdcTag.xyz)
			
			in coordsys local tagObject.rotation = mdcTag.GetRotation()
			
			tagObject
		),

		-- Load a surface and all its associated data (eg. animations, skins)
		fn LoadSurface=
		(
			local surface = MDC_Surface()
			surface.LoadFromStream bstream

			--surface.DebugOut()
			
			if not surface.IsValid() then
			(
				return false
			)
			
			-- Load triangles
			fseek bstream (surface.offsetStart + surface.offsetTriangles) #seek_set

			local modelTris = #()
			for i = 1 to surface.numTriangles do
			(
				local curTri = MDC_Triangle()
				curTri.LoadFromStream bstream
				
				append modelTris curTri.indices
			)
			
			-- Load shaders
			fseek bstream (surface.offsetStart + surface.offsetShaders) #seek_set
			local modelShaders = #()
			for i = 1 to surface.numShaders do
			(
				local curShader = MDC_Shader()
				curShader.LoadFromStream bstream
				append modelShaders curShader.name
			)
			
			-- Load texture co-ordinates
			fseek bstream (surface.offsetStart + surface.offsetTexCoords) #seek_set
			local modelTexCoords = #()
			for i = 1 to surface.numVertices do
			(
				local curTC = MDC_TexCoord()
				curTC.LoadFromStream bstream
				append modelTexCoords curTC.stu
			)
			
			-- Create a material
			local modelMaterial = standardmaterial name:(header.name + " material")
			
			local diffuseMap = texCache.FindTexture modelShaders[1]
			
			if diffuseMap != undefined then
			(
				modelMaterial.maps[2] = diffuseMap
				modelMaterial.mapEnables[2] = true
				showTextureMap modelMaterial diffuseMap true
			)
			
			-- Create mesh in gmax
			local gmaxMesh = mesh name:surface.name numVerts:surface.numVertices numFaces:surface.numTriangles \
									position:[0,0,0] scale:[1,1,1] material:modelMaterial
			
			-- Add faces
			setMesh gmaxMesh faces:modelTris
			
			-- Add per-vertex tex-coords and texture faces
			setNumTVerts gmaxMesh surface.numVertices
			for i = 1 to surface.numVertices do
			(
				setTVert gmaxMesh i modelTexCoords[i]

			)
			
			buildTVFaces gmaxMesh
			for i = 1 to surface.numTriangles do
			(
				setTVFace gmaxMesh i modelTris[i]
			)
			--update gmaxMesh
			
			-- try and help garbage collection
			modelTexCoords = undefined
			modelTris = undefined
			gc()
			
			-- Now, on to the really fun bit - animation!
			
			-- Load base frames table
			fseek bstream (surface.offsetStart + surface.offsetFrameBaseFrames) #seek_set
			local baseFrames = #()
			for i = 1 to header.numFrames do
			(
				append baseFrames (ReadShort bstream #unsigned)
			)
			
			-- Load compressed frames table
			fseek bstream (surface.offsetStart + surface.offsetFrameCompFrames) #seek_set
			local compFrames = #()
			for i = 1 to header.numFrames do
			(
				append compFrames (ReadShort bstream #unsigned)
			)
			
			-- Load first frame as base mesh (not animated)
			local lastBaseFrame = -1
			local baseFrameVerts = #()
			
			local mdcFrame = frames[1]
			
			-- Is first frame a base frame?
			if compFrames[1] == 0xffff then
			(
				-- Load new base frame, with no deltas applied
				local frameOffset = baseFrames[1] * surface.numVertices * MDC_Vertex.GetSize()
				fseek bstream (surface.offsetStart + surface.offsetVertices + frameOffset) #seek_set
				
				local curVertex = MDC_Vertex()
				for i = 1 to surface.numVertices do
				(
					curVertex.LoadFromStream bstream
					baseFrameVerts[i] = curVertex.position
				)
				
				-- Set base frame with no deltas
				setMesh gmaxMesh vertices:(for v in baseFrameVerts collect (v + frames[1].localOrigin))
			)
			else
			(
				-- First frame is a delta - unlikely, but we'll play along anyway
				
				-- Load base frame
				local frameOffset = baseFrames[1] * surface.numVertices * MDC_Vertex.GetSize()
				fseek bstream (surface.offsetStart + surface.offsetVertices + frameOffset) #seek_set
							
				local curVertex = MDC_Vertex()
				for i = 1 to surface.numVertices do
				(
					curVertex.LoadFromStream bstream
					baseFrameVerts[i] = curVertex.position
				)

				-- Load and add deltas onto base frame
				local deltaOffset = compFrames[1] * surface.numVertices * 4
				fseek bstream (surface.offsetStart + surface.offsetCompVerts + deltaOffset) #seek_set
				
				local unpackScale = 3.0 / 64.0

				local modelVerts = #()
				for i = 1 to surface.numVertices do
				(
					local cx = (ReadByte bstream #unsigned) - 127
					local cy = (ReadByte bstream #unsigned) - 127
					local cz = (ReadByte bstream #unsigned) - 127
					local cn = (ReadByte bstream #unsigned)
						
					local delta = [cx as float, cy as float, cz as float]
					delta *= 3.0
					delta /= 64.0
						
					local animVert = frames[1].localOrigin + baseFrameVerts[i] + delta -- ((delta - [127.0,127.0,127.0]) / 64.0)
					append modelVerts animVert
				)
				
				-- Set the first frame	
				setMesh gmaxMesh vertices:modelVerts
			)
			
			fseek bstream (surface.offsetStart + surface.offsetFrameCompFrames) #seek_set

			-- Construct each subsequent animation frame
			animate on
			(
				progressStart ("Loading " + surface.name + ", please wait...")
			
				for curFrame = 1 to header.numFrames do
				(
					local mdcFrame = frames[curFrame]
					--format "mdcFrame localorig = %\n" mdcFrame.localOrigin
					
					-- Is this a base frame?
					if compFrames[curFrame] == 0xffff then
					(
						-- Load new base frame, with no deltas applied
						local frameOffset = baseFrames[curFrame] * surface.numVertices * MDC_Vertex.GetSize()
						fseek bstream (surface.offsetStart + surface.offsetVertices + frameOffset) #seek_set
						
						local curVertex = MDC_Vertex()
						for i = 1 to surface.numVertices do
						(
							curVertex.LoadFromStream bstream
							baseFrameVerts[i] = curVertex.position
						)
						
						-- Set base frame with no deltas
						at time curFrame
						(
							for i = 1 to surface.numVertices do
							(
								meshop.SetVert gmaxMesh i (mdcFrame.localOrigin + baseFrameVerts[i])
							)
						)

					)
					else
					(
						-- Do we need to load a new base frame?
						if lastBaseFrame != baseFrames[curFrame] then
						(
							-- Load base frame
							local frameOffset = baseFrames[curFrame] * surface.numVertices * MDC_Vertex.GetSize()
							fseek bstream (surface.offsetStart + surface.offsetVertices + frameOffset) #seek_set
							
							local curVertex = MDC_Vertex()
							for i = 1 to surface.numVertices do
							(
								curVertex.LoadFromStream bstream
								baseFrameVerts[i] = curVertex.position
							)
						)

						-- Load and add deltas onto base frame
						local deltaOffset = compFrames[curFrame] * surface.numVertices * 4
						fseek bstream (surface.offsetStart + surface.offsetCompVerts + deltaOffset) #seek_set
						
						at time curFrame
						(
							local unpackScale = 3.0 / 64.0
						
							for i = 1 to surface.numVertices do
							(
								local cx = (ReadByte bstream #unsigned) - 127
								local cy = (ReadByte bstream #unsigned) - 127
								local cz = (ReadByte bstream #unsigned) - 127
								local cn = (ReadByte bstream #unsigned)
								
								local delta = [cx as float, cy as float, cz as float]
								delta *= 3.0
								delta /= 64.0
								
								if ( curFrame == 3) then
								(
									--format "delta = % \n" delta
								
								)
								
								local animVert = baseFrameVerts[i] + delta -- ((delta - [127.0,127.0,127.0]) / 64.0)
								
								meshop.SetVert gmaxMesh i (mdcFrame.localOrigin + baseFrameVerts[i] + delta)
							)
						)
					)
					
					update gmaxMesh
					
					lastBaseFrame = baseFrames[curFrame]
					
					progressUpdate (100.0 * curFrame / header.numFrames)
				)
				
				progressEnd()
			)
				
			-- Move to end of surface ready for the next one
			fseek bstream (surface.offsetStart + surface.offsetEnd) #seek_set
		
			-- Return gmax mesh
			gmaxMesh		
		),
		
	
		-- Load and construct the entire MDC model
		fn LoadModel=
		(
			-- Load MDC_Header
			fseek bstream 0 #seek_set
			
			header = MDC_Header()
			header.LoadFromStream bstream
			--header.DebugOut()
			
			if not header.IsValid() then
			(
				format "Invalid header, aborting.\n"
				return false;
			)
			
			-- Load frames
			fseek bstream header.offsetFrames #seek_set
			frames = #()
			
			local curFrame = MDC_Frame()
			for i = 1 to header.numFrames do
			(
				curFrame.LoadFromStream bstream
				
				append frames (copy curFrame)
			)

			-- Load tag names
			fseek bstream header.offsetTagNames #seek_set
			tagNames = #()
			tagObjects = #()
			for i = 1 to header.numTags do
			(
				local curTagName = MDC_TagName()
				curTagName.LoadFromStream bstream
				append tagNames curTagName.name
				
				--format "Loaded tag %\n" curTagName.name
			)
			
			
			-- Process tags only if we're loading or attaching
			if tagAction.state != 1 then
			(
				-- Load tag frames (BROKEN for more than one frame)
				fseek bstream header.offsetTags #seek_set
				tagFrames = #()
				
				local relativeTag
				
				for i = 1 to header.numFrames do
				(
					for j = 1 to header.numTags do
					(
						local curTag = MDC_TagFrame()
						curTag.LoadFromStream bstream
	
						--format "Frame %, %: pos = % rot = %\n" i tagNames[j] curTag.xyz curTag.angles
						
						append tagFrames curTag
						
						if i > 1 then
						(
							animate on
							(
								at time i
								(
									in coordsys world 
									(
										tagObjects[j].transform = curTag.GetMatrix()
									)
								)
							)
						)
						else
						(
							if curTag.isRelative then
							(
								relativeTag = j
							)
						
							local tagObject = CreateTagObject tagNames[j] curTag
							append tagObjects tagObject
						
							in coordsys world
							(
								tagObject.transform = curTag.GetMatrix()
							)
							
							animate on
							(
								at time 1
								(
								in coordsys world
								(
									tagObject.transform = curTag.GetMatrix()
								)
								)
							)
						)
					)
				)
			)			
		
			
			-- Load surfaces
			surfaceList = #()
		
			fseek bstream header.offsetSurfaces #seek_set

			for i = 1 to header.numSurfaces do
			(
				-- Load and create the surface model
				local newObject = LoadSurface()
				append surfaceList newObject
			)
			
			-- Auto attach to scene tags, if we can find 'em
			if tagAction.state == 3 then
			(
				if relativeTag != undefined then
				( 
				
					-- Delete our tags
					-- delete tagObjects
				
					local targetTag
					
					-- Find target tag
					for obj in objects do
					(
						if obj.name == tagNames[relativeTag] then
						(
							if (findItem tagObjects obj) == 0 then
							(
								targetTag = obj
							)
							--format "Checked out %\n" obj
						)
					)
					
					if targetTag == undefined then
					(
						format "Warning: Can't find parent tag % to attach to\n" tagNames[relativeTag]
					)
					else
					(
						-- Attach all our surfaces to this tag
						for i = 1 to header.numSurfaces do
						(
							attachObjects targetTag surfaceList[i]
							in coordsys parent surfaceList[i].rotation = (tagFrames[relativeTag].GetRotation())
						)
					)
				)
				
				-- Delete tags so we don't confuse ourselves
				delete tagObjects
			)
									
			
			-- Set animation slider range and FPS
			if header.numFrames > 1 then
			(
				animationRange = interval 0 header.numFrames -- Hewster was 1
				frameRate = 30.0 -- Hewster was 10
			)
			
			-- All done, try and clean up
			texCache.Clear()
			
			gc()
			true
		)
	)
	
	
	-------------------
	-- Final UI Bits --
	-------------------
	
	on pakDirEdit entered newText do
	(
		g_gameDataPath = newText
	)
	
 	on setPakRootButton pressed do
	(
		local newPakDir = getSavePath caption:"Set Game Data Path"
		if newPakDir != undefined do
		(
			g_gameDataPath = newPakDir
			UpdateUI()
		)
	)
	
	on importButton pressed do
	(
		local modelFilename = getOpenFileName caption:"Import Wolfenstein Model" filename:(GetMostRecentPath()) \
					types:"Return to Castle Wolfenstein MDC (*.mdc)|*.mdc|All Files (*.*)|*.*|"
		


		if modelFilename != undefined then
		(
			if DoesFileReallyExist modelFilename then
			(
				AddRecentFile modelFilename
			
				--try
				(
					local loader = MDC_Loader()
					loader.Open modelFilename
					loader.LoadModel()
					loader.Close()
				)
				--catch
				(
					--throw
					--FatalError "Could 
				)
				
				max views redraw
			)
		)
	)
)

